: \
  #TIB @ >IN !
; IMMEDIATE
\ Use the new word: A backslash starts a comment up the end of the line.

\ An opening paren starts a comment up a closing paren in the same line.
\
: (
  29 WORD DROP
; IMMEDIATE

\ Switch numeric base to hexadecimal.
\
: HEX ( -- )
  16 BASE !
;

\ Switch numeric base to decimal.
\
: DECIMAL ( -- )
  10 BASE !
;

\ Logical shift left.
\
: LSHIFT ( n bits -- n' )
  (SHIFT)
;

\ Logical shift right.
\
: RSHIFT ( n bits -- n' )
  NEGATE (SHIFT)
;

\ Return the execution token of the next word.
\
: ' ( -- xt )
  BL WORD FIND DROP             \ xt
;

\ Compile the following word, which is usually not immediate, because
\ otherwise the first thing it compiled would be used instead of itself.
\ The return address of the code calling us points to the word to be
\ compiled, so DUP it, increment the return address and compile the DUPed
\ value.
\
: COMPILE ( -- )
  R> DUP CELL+ >R @ ,
;

\ Compile the following word, no matter if it is immediate or not.
\
: [COMPILE] ( -- )
  ' ,
; IMMEDIATE

\ Compile code that returns the execution token of word following ['],
\ even if it is immediate.  ['] consists of tick and LITERAL,
\ the latter being immediate, hence we compile it using [COMPILE].
\
: ['] ( "name" -- )
  ' [COMPILE] LITERAL
; IMMEDIATE

\ Compile a new word with a DOES> handler that calls EXIT.  It is tempting
\ to compile EXIT as DOES> handler, but EXIT is written in machine code
\ and you can not do a Forth branch to machine code.
\
: CREATE ( "name" -- )
  (CREATE)
  COMPILE DODOES
  COMPILE (EXIT)
;

\ Turn data word into some kind of object that has only a single
\ method: The code following DOES>.  The handler is part of the
\ word using DOES>, not compiled into the CREATEd word.  That's
\ why the data start of the word is passed to it.
\
: DOES> ( -- addr )
  LATEST @ LFA>CFA              \ cfa
  CELL+ R>                      \ cfa+1 code-address
  SWAP !                        \ cfa
;

\ Define a single-cell variable.
\
: VARIABLE ( -- )
  CREATE 0 ,
;

\ Constant definition
\
: CONSTANT ( value -- )
  CREATE , DOES> @
;

\ Recurse to the currently defined word.
\
: RECURSE ( -- )
  LATEST @                      \ curlfa
  LFA>CFA                       \ xt
  ,
; IMMEDIATE 

\ Clear stack and QUIT
\
: ABORT ( i*x --    R: j*x -- )
  SP0 @ SP! QUIT
;


\ Conditional execution
\
\ The structure is:
\
\   flag IF ... THEN
\   flag IF ... ELSE ... THEN
\
\ The condition is used by IF, which marks the conditionally executed
\ part.  THEN terminates the parts.  Do not think of it as "if condition
\ then do something" but as "if condition do something, then the program
\ continues".
\
\ The tags can be used to check proper nesting of control structures
\ while compiling them.  Here, it makes sure THEN follows IF or ELSE and
\ ELSE can not be used twice.
\
: IF ( compilation: -- ifnotaddr 0 )
  COMPILE 0BRANCH               \ compile 0BRANCH
  HERE                          \ get address of target
  0 ,                           \ reserve space for target
  0                             \ push tag 0
; IMMEDIATE

: THEN ( compilation: ifnotaddr|elsedoneaddr 0|1 -- )
  DROP                          \ pop tag 0 or 1
  HERE                          \ get current address
  SWAP !                        \ store as branch target
; IMMEDIATE

: ELSE ( compilation: ifnotaddr 0 -- elsedoneaddr 1 )
  DROP                          \ pop tag 0
  COMPILE BRANCH                \ definite branch to just over the false-part
  HERE                          \ get address of target
  0 ,                           \ reserve space for target
  SWAP                          \ now back-fill the original (IF) offset
  1                             \ push tag 1
  [COMPILE] THEN
  1                             \ push tag 1 for final THEN
; IMMEDIATE


\ Simple loops
\
\ The structure is:
\
\   BEGIN ... AGAIN
\   BEGIN ... flag UNTIL
\   BEGIN ... flag WHILE ... REPEAT
\
\ Start a simple loop.  Remember the current address to use it later as
\ branch target.
\
: BEGIN ( compilation: -- beginaddr 2 )
  HERE   
  2                             \ push tag 2
; IMMEDIATE

\ Unconditional loop
\
: AGAIN ( compilation: adrs 2 -- )
  DROP                          \ pop tag 2
  COMPILE BRANCH ,
; IMMEDIATE

\ Post checked loop.  First compile a conditional branch, then append
\ the target address from the stack using comma.
\
: UNTIL ( compilation: addr 2 -- )
  DROP                          \ pop tag 2, should check for error here
  COMPILE 0BRANCH ,
; IMMEDIATE

\ Pre checked loop. WHILE is in fact a BEGIN/IF/AGAIN/THEN sequence, so we manipulate
\ the tags accordingly and compile those words, letting them compile branches and resolve
\ their targets.
\
: WHILE ( compilation: beginaddr 2 -- beginaddr whilenotaddr 3 )
  DROP                          \ pop tag 2
  [COMPILE] IF DROP             \ conditionally leave loop
  3
; IMMEDIATE

: REPEAT ( compilation: beginaddr whilenotaddr 3 -- )
  DROP                          \ pop tag 3
  SWAP
  2 [COMPILE] AGAIN             \ branch to BEGIN
  0 [COMPILE] THEN              \ resolve WHILE branch
; IMMEDIATE


\ Counted loops
\
\ The odd thing about loops is that leaving them from inside their body breaks
\ the control structure hierarchy.  Control structures are usually nested, but
\ LEAVE may be located inside an if statement or anywhere else, ignoring
\ all nesting rules but being somewhere in the DO body.
\
\ For that reason, LEAVE words can not be managed using the control flow
\ stack.  Some implementations use a separate leave stack that contains
\ entries for each LEAVE, separated by a delimiter for DO, others (like this
\ one) store this information in the compiled code by chaining the LEAVE and
\ DO cells to a linked list, using a loop pointer (LP) that points to its head.
\
\ This example shows how things work in detail:
\
\ 10 0 DO
\   2 1 DO
\     1 IF LEAVE THEN
\     LEAVE
\   LOOP
\   1 IF LEAVE THEN
\ LOOP
\
\ It will be compiled to:
\
\          LIT 10              10
\          LIT 0                0
\ a        (DO)                DO
\          LIT 2   <-----+       2
\          LIT 1         |       1
\ b        (DO)          |       DO
\          LIT 1   <---+ |         1
\          0BRANCH --+ | |         IF
\ c        d         | | |
\   +----- BRANCH    | | |           LEAVE
\   |      f         | | |
\ d |  +-- BRANCH  <-+ | |         THEN LEAVE
\ e |  |   f           | |
\   |  |   (LOOP)      | |       LOOP
\   |  |   b+1 --------+ |
\ f +->+-> 2RDROP        |
\          LIT 1         |       1
\          0BRANCH --+   |       IF
\ g        h         |   |
\      +-- BRANCH    |   |         LEAVE
\      |   i         |   |
\ h    |   (LOOP)  <-+   |       THEN LOOP
\      |   a+1 ----------+
\ i    +-> 2RDROP
\
\ This is the data stack, DA content and compiled code during compilation of
\ the above program:
\
\ Code                 Stack   DA   Compiled to      Action
\                              0
\ 10 0 DO              a       a    LIT 10 LIT 0 0   LP = a
\ 2 1 DO               a b     b    LIT 2 LIT 1 a    LP = b
\ 1 IF                 a b c   b    LIT 1 BRANCH0 0
\ LEAVE                a b c   b    BRANCH b         LP = d
\ THEN                 a b     b                     store branch target d in c
\ LEAVE                a b     b    BRANCH d         LP = e
\ LOOP                 a       a    (LOOP) b+1       LP = d
\                                   2RDROP           store leave target f in e
\                                                    LP = b
\                                                    store leave target f in d
\                                                    LP = a
\                                                    store (DO) in b
\ 1 IF                 a g     a    LIT 1 BRANCH0 0
\ LEAVE                a g     a    BRANCH a         LP = h
\ THEN                 a       a                     store branch target h in g
\ LOOP                         0    (LOOP) a+1       LP = a
\                                                    store leave target i in h
\                                                    LP = 0
\                                                    store (DO) in a
\     
\ DO compiles the contents of LP as cell to the previous DO cell and updates
\ LP with the address of the compiled cell, putting it at the head of the
\ linked list.  Further, puts this address and a tag to the control flow stack.
\
\ LEAVE compiles BRANCH and a cell containing the contents of LP, updating LP
\ to the address of this cell.
\
\ LOOP takes the DO tag off the stack, compiles (LOOP) and
\ converts the linked list up to the matching DO into actual code.  While LP
\ is not equal the DO address on the stack, it saves the contents of the
\ cell LP points to, stores the current address there, and sets LP to the
\ saved contents.  After filling all LEAVE branch targets, it saves the
\ contents of the cell LP points to, stores (DO) there and sets LP to the
\ saved contents.  Finally, it compiles 2RDROP, which is also the branch
\ target for LEAVE.
\
\ At runtime, the only thing to watch is manipulating the return stack.
\ (DO) can't simply push the limit and index to the return stack, because
\ it would bury the caller return address underneath.  Instead, it has
\ to get the return address, then move the limit and index from data to
\ return stack, and keep the caller return address on top.  (LOOP) does
\ not only care not to destroy the return address, it manipulates it to
\ use EXIT as indirect branch to the beginning of the loop or to skip the
\ jump target following (LOOP) to end the loop.
\
VARIABLE LP

: DO ( compilation: -- adrs 4  runtime: end index --   R: -- end index )
  HERE                          \ adrs
  LP @ ,                        \ adrs
  DUP LP !                      \ adrs
  4                             \ adrs 4
; IMMEDIATE

: (DO)          ( end index --   R: return -- end index return )
  SWAP                          \ index end   R: return
  R> SWAP >R                    \ index return   R: end
  SWAP >R                       \ return   R: end index
  >R                            \    R: end index return
; 

: LEAVE ( compilation: adrs 4 -- adrs 4  runtime: --  R: end index -- )
  ['] BRANCH ,                  \
  HERE                          \ adrs
  LP @ ,                        \ adrs
  LP !                          \
; IMMEDIATE

: (LOOP) ( --  R: end index return -- end index+1 return )
  R>
  R> 1+ >R 2R@ =
  IF CELL+ ELSE @ THEN
  >R
;

: LOOP ( compilation: adrs 4 --  runtime: --   R: end index -- | end index+1 )
  DROP                          \ adrs
  COMPILE (LOOP)                \ adrs
  DUP CELL+ ,                   \ adrs
  LP @                          \ adrs lp
  BEGIN                         \ adrs lp
    2DUP -                      \ adrs lp different?
  WHILE                         \ adrs lp
    DUP @ >R                    \ adrs lp  R: prev
    HERE SWAP !                 \ adrs  R: prev
    R>                          \ adrs prev
  REPEAT                        \ adrs prev
  @ >R                          \ adrs  R: prev
  ['] (DO) SWAP !                 \ R: prev
  R> LP !                       \
  COMPILE 2RDROP                \
; IMMEDIATE

: I ( -- i   R: i -- i )
  COMPILE R@
; IMMEDIATE


\ Number conversion
\
\ The HOLD buffer is a small memory area that is filled backwards, from
\ its end to its beginning.  The HOLD pointer (HP) points to the first
\ buffered character.  If the buffer is empty, it points to the first element
\ after the buffer.  It is convenient to store HP after the buffer,
\ because that allows to initialise it with itself, not having to bind the
\ HOLD buffer to a name.

32 ALLOT HERE 0 , : HP LITERAL ;

\ Add character to output buffer
\
: HOLD ( char -- )
  -1 CHARS HP +!
  HP @ C!
;

\ Begin conversion
\
: <# ( -- )
  HP HP !
;

\ Convert one digit
\
: # ( u1 -- u2 )
  BASE @                        \ u base
  /MOD                          \ remainder quotient
  SWAP                          \ quotient remainder
  >DIGIT HOLD                   \ quotient
;

\ Convert remaining digits
\
: #S ( u1 -- u2 )
  BEGIN                         \ u1
    # DUP 0=                    \ u1' u1-null?
  UNTIL
;

\ End conversion and generate string
\
: #> ( n -- str len )
  DROP
  HP @ HP                       \ holdptr holdptrptr
  OVER                          \ holdptr holdptrptr holdptr
  -                             \ holdptr len
;

\ Remove sign
\
: ABS ( n -- u )
  DUP 0< IF NEGATE THEN
;

\ Add minus sign if n is negative
\
: SIGN ( n -- )
  0< IF 45 HOLD THEN
;

\ Display n signed
\
: . ( n -- )
  <# DUP ABS #S SWAP SIGN #> TYPE SPACE ;
;

\ Display n unsigned
\
: U. ( n -- )
  <# #S #> TYPE SPACE ;
;

\ Fetch double cell
\
: 2@ ( addr -- n2 n1 )
  DUP CELL+                     \ addr addr+1
  @                             \ addr n2
  SWAP @                        \ n2 n1
;

\ Build a new wordlist
\
: WORDLIST ( -- wid )  
  HERE 32 CELLS ALLOT
;
